#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  4 21:16:48 2019

@author: peruzzetto
"""

import shapely.geometry as geom
import shapely
import numpy as np
import matplotlib.pyplot as plt
import warnings
import copy
import scipy.interpolate

# class Extent():
#    def __init__(self,coords=None,pos=None):
#        self.coords=coords
#        self.pos=pos


class SectionPoints:
    def __init__(self, *args, normalized=False, z=[]):
        self.normalized = normalized
        self.pos = []
        self.points = []
        self.z = []
        try:
            if len(args) == 2:
                linear_geom = args[0]
                self.line = linear_geom
                points = args[1]
                if isinstance(points, geom.LineString):
                    points = geom.MultiPoint(points.coords[:])

                if isinstance(points, geom.MultiPoint):
                    if not points.is_empty:
                        self.points = points
                        self.pos = [
                            linear_geom.project(point, normalized=normalized)
                            for point in points.geoms
                        ]
                        self.z = z
                elif isinstance(points, list):
                    self.pos = points
                    self.points = geom.MultiPoint(
                        [
                            linear_geom.interpolate(
                                s, normalized=self.normalized
                            )
                            for s in points
                        ]
                    )
                    self.z = z
                else:
                    warnings.warn(
                        "Second argument must be either shapely MultiPoint, LinearString or float list. \
                                  SectionPoints is created as empty"
                    )
                    self.set_empty()

                self.order_from_pos()

            else:
                try:
                    assert (
                        len(args) == 0
                    ), "0 or 2 arguments must be given for SectionPoints creation"
                    self.set_empty()
                except AssertionError:
                    raise
        except AssertionError:
            raise

    @property
    def z(self):
        return self._z

    @z.setter
    def z(self, z):
        assert isinstance(z, list), "z must be a float list"
        assert all(
            type(zz) in [float, int, np.float32, np.float64] for zz in z
        ), "z contains non int or float elements"
        if isinstance(self.pos, list):
            #            print(self.pos,z)
            assert (len(z) == 0) | (
                len(z) == len(self.pos)
            ), "length of z does not match points input"
        self._z = z

    def set_empty(self):
        self.points = geom.MultiPoint()
        self.pos = []
        self.z = []

    def order_from_pos(self):
        if len(self.pos) > 1:
            ind = sorted(range(len(self.pos)), key=lambda k: self.pos[k])
            self.points = geom.MultiPoint([self.points.geoms[i] for i in ind])
            self.pos = [self.pos[i] for i in ind]
            if len(self.z) > 1:
                self.z = [self.z[i] for i in ind]

    def filter_by_pos(self, posmin, posmax):
        tmp = np.array(self.pos)
        ind = np.argwhere((tmp > posmin) & (tmp < posmax))
        if len(ind) > 0:
            self.points = geom.MultiPoint(
                [self.points.geoms[i[0]] for i in ind]
            )
            self.pos = [self.pos[i[0]] for i in ind]
            if len(self.z) > 1:
                self.z = [self.z[i[0]] for i in ind]
        else:
            self.points = geom.MultiPoint()
            self.pos = []
            self.z = []

    def add_points(self, points, z=[]):
        if not isinstance(points, list):
            points = [points]

        points2 = []
        pos2 = []
        z2 = []

        for point in points:
            if type(point) in [float, int, np.float32, np.float64]:
                pos2.append(point)
                points2.append(
                    geom.Point(
                        self.line.interpolate(
                            point, normalized=self.normalized
                        )
                    )
                )
            elif point.isinstance(SectionPoints):
                try:
                    assert (
                        point.line == self.line
                    ), "Combining SectionPoints along different lines"
                    pos2 = pos2 + point.pos
                    points2 = points2 + [pp for pp in point.points]
                    z2 = z2 + point.z
                except AssertionError:
                    raise
            else:
                if isinstance(point, geom.Point):
                    point = [point]
                points2 = points2 + [pp for pp in point]
                pos2 = pos2 + [
                    self.line.project(pp, normalized=self.normalized)
                    for pp in point
                ]
        self.pos = self.pos + pos2
        self.points = geom.MultiPoint([pp for pp in self.points] + points2)
        if len(z2) > 0:
            self.z = self.z + z2
        else:
            self.z = self.z + z
        self.order_from_pos()


class Section:
    def __init__(
        self,
        points,
        name,
        orientation=None,
        direction="",
        normalized=False,
        topography=None,
        name_sep="-",
        fz_old=None,
        control_points=None,
        dinterp=10,
        interp_spline_options={},
        line_points_as_control_points=True,
        point_name_dist=None,
    ):
        if isinstance(points, geom.linestring.LineString):
            self.line = points
        elif isinstance(points, list):
            self.line = geom.LineString(points)
        elif isinstance(points, np.ndarray):
            tmp = [
                (points[i, 0], points[i, 1]) for i in range(points.shape[0])
            ]
            self.line = geom.LineString(tmp)

        self.name = name
        self.direction = direction
        self.normalized = normalized
        self.topography = topography
        self.fz_old = fz_old
        self.interp_spline = None
        self.interp_spline_options = interp_spline_options
        self.control_points = control_points
        self.intersection_points = {}
        self.orientation = orientation
        self.extent = None
        self.extreme_points_names = []
        self.name_sep = name_sep
        self.point_name_dist = point_name_dist
        self.d_interp = dinterp
        self.interp_points = SectionPoints()

        self.set_extreme_points_names()

        if (control_points is None) & line_points_as_control_points:
            self.control_points = SectionPoints(
                self.line, self.line, normalized=self.normalized
            )

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, orientation):
        self._orientation = orientation
        if isinstance(orientation, str):
            if orientation == "S-N":
                reverse = self.line.coords[0][1] > self.line.coords[-1][1]
            elif orientation == "N-S":
                reverse = self.line.coords[0][1] < self.line.coords[-1][1]
            elif orientation == "W-E":
                reverse = self.line.coords[0][0] > self.line.coords[-1][0]
            elif orientation == "E-W":
                reverse = self.line.coords[0][0] < self.line.coords[-1][0]
            else:
                warnings.warn(
                    "orientation not set, must be either 'S-N', 'N-S', 'W-E' or 'E-W'"
                )
            if reverse:
                self.line = shapely.ops.substring(
                    self.line, self.line.length, 0
                )

    def set_extreme_points_names(self, sep=None):
        if sep is None:
            sep = self.name_sep
        if sep != "":
            names = self.name.split(self.name_sep)
            if len(names) == 2:
                self.extreme_points_names = names
            else:
                warnings.warn(
                    "Extreme points names could not be set, invalid string separator"
                )

    def set_extent(self, other):
        if other is not None:
            intersections = self.line.intersection(other)
            self.extent = SectionPoints(
                self.line, intersections, normalized=self.normalized
            )
            self.filter_control_points()
            if self.fz_old is not None and len(self.extent.pos) > 0:
                coords = np.array(
                    geom.LineString(self.extent.points.geoms).coords
                )
                self.extent.z = list(
                    self.fz_old(coords[:, 0], coords[:, 1], grid=False)
                )

    def filter_control_points(self):
        if self.extent is not None:
            if (self.control_points is not None) & (len(self.extent.pos) > 1):
                self.control_points.filter_by_pos(
                    self.extent.pos[0], self.extent.pos[-1]
                )
            if len(self.extent.pos) == 0:
                self.control_points.set_empty()

    def filter_intersection_points(self):
        if self.extent is not None:
            for key in self.intersection_points:
                if (self.intersection_points[key] is not None) & (
                    len(self.extent.pos) > 1
                ):
                    self.intersection_points[key].filter_by_pos(
                        self.extent.pos[0], self.extent.pos[-1]
                    )
                if len(self.extent.pos) == 0:
                    self.intersection_points[key].set_empty()

    def set_control_points(self, pos, z):
        try:
            assert len(pos) == len(z), "pos and z must have same length"
            self.control_points = SectionPoints(
                self.line, pos, z=z, normalized=self.normalized
            )
            self.filter_control_points()
        except AssertionError:
            raise

    def set_intersection_points(self, name2, pos, z=[]):
        try:
            assert len(pos) == len(z), "pos and z must have same length"
            self.intersection_points[name2] = SectionPoints(
                self.line, pos, z=z, normalized=self.normalized
            )
            self.filter_intersection_points()
        except AssertionError:
            raise

    def set_interp_spline(self):
        ss = copy.copy(self.control_points.pos)
        zz = copy.copy(self.control_points.z)
        for key in self.intersection_points:
            ss = ss + copy.copy(self.intersection_points[key].pos)
            zz = zz + copy.copy(self.intersection_points[key].z)
        ind = sorted(range(len(ss)), key=lambda k: ss[k])
        ss.sort()
        zz = [zz[i] for i in ind]
        if len(ss) > 0:
            if self.extent is not None:
                if (len(self.extent.pos) > 0) & (len(self.extent.z) > 0):
                    ss.append(self.extent.pos[-1])
                    zz.append(self.extent.z[-1])
                    ss.insert(0, self.extent.pos[0])
                    zz.insert(0, self.extent.z[0])
            npts = len(ss)
            self.interp_spline = scipy.interpolate.splrep(
                ss, zz, s=0, k=min(3, npts - 1), **self.interp_spline_options
            )

    def set_interp_points(self):
        if len(self.interp_points.pos) > 0 or len(self.extent.pos) > 1:
            if self.extent is not None and len(self.extent.pos) > 1:
                #                print(self.name,len(self.extent.pos))
                pos1 = self.extent.pos[0]
                pos2 = self.extent.pos[-1]
            else:
                pos1 = 0
                pos2 = self.line.length
            pos = list(
                np.arange(pos1, pos2 + self.d_interp / 2, self.d_interp)
            )
            if pos[-1] != pos2:
                pos.append(pos2)
            if self.interp_spline is None:
                z = []
            else:
                z = scipy.interpolate.splev(pos, self.interp_spline, der=0)
            self.interp_points = SectionPoints(self.line, pos, z=list(z))
        else:
            self.interp_points = SectionPoints()

    def plot(
        self,
        axe=None,
        figsize=None,
        npos=200,
        old_topo_properties=None,
        new_topo_properties=None,
        extent_properties=None,
        control_points_properties=None,
        fz_new=None,
        comp_surf=None,
        plot_old_topo=True,
        fz_old=None,
    ):
        if axe is None:
            fig, axe = plt.subplots(1, 1, figsize=figsize)
        else:
            fig = axe.figure

        if old_topo_properties is None:
            old_topo_properties = {"lw": 2, "color": "k", "linestyle": "-"}
        if new_topo_properties is None:
            new_topo_properties = {
                "lw": 2,
                "color": [0.5, 0.5, 0.5],
                "linestyle": "-",
            }

        if extent_properties is None:
            extent_properties = {
                "linestyle": "None",
                "marker": "o",
                "ms": 6,
                "mec": "black",
                "mfc": "white",
            }

        if control_points_properties is None:
            control_points_properties = {
                "linestyle": "None",
                "marker": "o",
                "ms": 6,
                "mec": "black",
                "mfc": [0.5, 0.5, 0.5],
            }

        pos = np.linspace(0, self.line.length, npos)
        pts = geom.LineString([self.line.interpolate(p) for p in pos])
        coords = np.array(pts.coords)

        if self.extent is not None and len(self.extent.pos) > 0:
            ind = np.argwhere(
                (pos >= self.extent.pos[0]) & (pos <= self.extent.pos[1])
            )
            ind = np.squeeze(ind)
        elif len(self.control_points.pos) > 1:
            ind = np.argwhere(
                (pos >= self.control_points.pos[0])
                & (pos <= self.control_points.pos[1])
            )
            ind = np.squeeze(ind)
        else:
            ind = []

        if fz_old is None:
            fz_old = self.fz_old

        if fz_old is not None:
            old_topo = fz_old(coords[:, 0], coords[:, 1], grid=False)
        else:
            old_topo = None

        if len(ind) > 0:

            def evaluate_new_topo(*args):
                if fz_new is None:
                    out = scipy.interpolate.splev(
                        pos[ind], self.interp_spline, der=0
                    )
                else:
                    out = fz_new(
                        coords[ind[0] : ind[-1] + 1, 0],
                        coords[ind[0] : ind[-1] + 1, 1],
                    )
                    if old_topo is not None:
                        if comp_surf == "min":
                            out = np.minimum(out, old_topo[ind])
                        elif comp_surf == "max":
                            out = np.maximum(out, old_topo[ind])
                return out

            if plot_old_topo:
                axe.plot(pos, old_topo, **old_topo_properties)
                new_topo = np.copy(old_topo)
                new_topo[ind] = evaluate_new_topo(old_topo)
                pos_new = pos
            else:
                new_topo = evaluate_new_topo()
                pos_new = pos[ind]

            axe.plot(pos_new, new_topo, **new_topo_properties)

            # Plot extent and control points
            if self.extent is not None:
                if len(self.extent.pos) > 1:
                    for i in [0, -1]:
                        axe.plot(
                            self.extent.pos[i],
                            self.extent.z[i],
                            **extent_properties,
                        )

            for i in range(len(self.control_points.points.geoms)):
                axe.plot(
                    self.control_points.pos[i],
                    self.control_points.z[i],
                    **control_points_properties,
                )

            for key in self.intersection_points:
                for i in range(
                    len(self.intersection_points[key].points.geoms)
                ):
                    axe.plot(
                        self.intersection_points[key].pos[i],
                        self.intersection_points[key].z[i],
                        **control_points_properties,
                    )

            axe.set_aspect("equal")
            axe.set_xlabel("Distance (m)")
            axe.set_ylabel("Altitude (m)")
            axe.set_title(self.name)
            axe.grid("both")

            fig.tight_layout()

        return fig, axe

    def plot_on_map(
        self,
        axe=None,
        figsize=None,
        point_name_dist=None,
        npts=2,
        section_properties=None,
        text_properties=None,
        control_points_properties=None,
    ):
        xx, yy = self.line.coords.xy
        if control_points_properties is None:
            control_points_properties = {
                "linestyle": "None",
                "marker": "o",
                "ms": 5,
                "color": "k",
                "mfc": "k",
            }

        if axe is None:
            fig, axe = plt.subplots(1, 1, figsize=figsize)
        # else:
        #     fig = axe.figure

        if section_properties is None:
            section_properties = dict()
        axe.plot(xx, yy, **section_properties)
        if len(self.control_points.points.geoms) > 0:
            for p in self.control_points.points.geoms:
                axe.plot(
                    p.coords[0][0], p.coords[0][1], **control_points_properties
                )

        for key in self.intersection_points:
            if len(self.intersection_points[key].points.geoms) > 0:
                for p in self.intersection_points[key].points.geoms:
                    axe.plot(
                        p.coords[0][0],
                        p.coords[0][1],
                        **control_points_properties,
                    )

        if len(self.extreme_points_names) > 0:
            if point_name_dist is None:
                point_name_dist = self.point_name_dist

            if point_name_dist is None:
                point_name_dist = self.line.length / 15

            if text_properties is None:
                text_properties = dict()

            for i in [0, 1]:
                xtmp = [xx[int((1 - 2 * npts) * i + npts - 1)], xx[-i]]
                ytmp = [yy[int((1 - 2 * npts) * i + npts - 1)], yy[-i]]

                vecx = xtmp[1] - xtmp[0]
                vecy = ytmp[1] - ytmp[0]
                nn = np.sqrt(vecx**2 + vecy**2)
                vecx = vecx / nn
                vecy = vecy / nn

                axe.text(
                    xx[-i] + point_name_dist * vecx,
                    yy[-i] + point_name_dist * vecy,
                    self.extreme_points_names[i],
                    horizontalalignment="center",
                    verticalalignment="center",
                    **text_properties,
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct  3 13:55:55 2019

@author: peruzzetto
"""

import shapely.geometry as geom
import digdem.plot
import numpy as np
import matplotlib.pyplot as plt
import os
import geopandas as gpd
import warnings
import logging
from collections import defaultdict
from digdem.section import Section
import scipy.interpolate
import itertools
from scipy.signal import convolve2d


class Tree(defaultdict):
    def __init__(self, value=None):
        super(Tree, self).__init__(Tree)
        self.value = value


class InterpPoints:
    def __init__(self, *args):
        if len(args) == 2:
            self.points = args[0]
            self.z = args[1]
        elif len(args) == 3:
            self.z = args[2]
            assert len(args[0]) == len(args[1])
            points = [[args[0][i], args[1][i]] for i in range(len(args[0]))]
            self.points = geom.MultiPoint(points)
        else:
            self.points = geom.MultiPoint()
            self.z = []
        assert len(self.points.geoms) == len(self.z)

    @property
    def xy(self):
        return np.array(geom.LineString(self.points.geoms).coords)

    def add(self, *args):
        for i in range(len(args)):
            if not isinstance(args[i], list):
                args[i] = [args[i]]
        if len(args) == 3:
            x = args[0]
            y = args[1]
            z = args[2]
            points = [geom.Point([x[i], y[i]]) for i in range(len(x))]
        elif len[args] == 2:
            z = args[1]
            points = args[0]
        self.z = self.z + z
        self.points = geom.MultiPoint(list(self.points) + points)

    def unique(self):
        xy = self.xy
        xy, ind = np.unique(xy, axis=0, return_index=True)
        self.z = [self.z[i] for i in ind]
        self.points = geom.MultiPoint(xy)

    def __len__(self):
        return len(self.points.geoms)


class SurfMod(dict):
    @staticmethod
    def read_shapefile(
        shapefile, direction_field_name="dir", orientation_field_name="ori"
    ):
        sections = []
        try:
            assert shapefile[-4:] == ".shp"
            shp_reader = gpd.read_file(shapefile)
            # print(shapefile)
            for row in shp_reader.index:
                line = shp_reader.loc[row, "geometry"]
                npoints = line.GetPointCount()
                points = []
                for i in range(npoints):
                    points.append(line.GetPoint(i))
                direction = None
                orientation = None
                if direction_field_name in shp_reader.columns:
                    direction = shp_reader.loc[row, direction_field_name]
                if orientation_field_name in shp_reader.columns:
                    orientation = shp_reader.loc[row, orientation_field_name]
                name = shp_reader.loc[row, "name"]
                sections.append(
                    Section(
                        points,
                        name,
                        direction=direction,
                        orientation=orientation,
                    )
                )
        except AssertionError:
            print("Format of " + shapefile + " file not recognized")
            exit(1)

        return sections

    @staticmethod
    def read_control_points_from_textfile(textfile, sep=" "):
        pos = {}
        z = {}
        z_intersections = Tree()
        with open(textfile, "r") as file:
            line = file.readline()
            try:
                assert (
                    line[0] == "#"
                ), "File must begin with a section name, identified with #"
                while line:
                    if line[-1] == "\n":
                        line = line[:-1]
                    if line[0] == "#":
                        key = line[1:]
                        pos[key] = []
                        z[key] = []
                    else:
                        vals = line.split(sep)
                        try:
                            if vals[0][0].isdigit():
                                pos[key].append(float(vals[0]))
                                z[key].append(float(vals[1]))
                            else:
                                z_intersections[key][vals[0]] = float(vals[1])
                        except ValueError:
                            pass
                    line = file.readline()
            except AssertionError:
                raise
        return pos, z, z_intersections

    def __init__(
        self,
        surf_old,
        mask,
        xaxis=None,
        yaxis=None,
        indexing="ij",
        sections=[],
        contour=None,
        normalized=False,
        sparse_interp=1,
        dinterp=10,
        ndv=-9999.9,
        smooth=0.5,
        surf_max=None,
        surf_min=None,
    ):
        self.logger = logging.getLogger(__name__)

        try:
            assert indexing in ["raster", "ij"], (
                "indexing must be " "raster" " or " "ij" ""
            )
            self.indexing = indexing
            if indexing == "raster":
                surf_old = np.flip(surf_old, axis=1).T
                mask = np.flip(mask, axis=1).T
            elif indexing == "ij":
                pass

            if xaxis is None:
                xaxis = np.arange(0, surf_old.shape[0], 1)
            else:
                assert type(xaxis) == np.ndarray, "xaxis is not np.ndarray"
            if yaxis is None:
                yaxis = np.arange(0, surf_old.shape[1], 1)
            else:
                assert type(yaxis) == np.ndarray, "yaxis is not np.ndarray"

            sizex_coherent = (
                mask.shape[0] == surf_old.shape[0] == xaxis.shape[0]
            )
            sizey_coherent = (
                mask.shape[1] == surf_old.shape[1] == yaxis.shape[0]
            )
            assert sizex_coherent & sizey_coherent, "Uncoherent shapes"

            self.surf_old = np.copy(surf_old)
            self.surf_old[np.isnan(self.surf_old)] = ndv
            self.surf_new = np.copy(self.surf_old)
            self.xaxis = xaxis
            self.yaxis = yaxis
            self.dx = np.abs(xaxis[1] - xaxis[0])
            self.dy = np.abs(yaxis[1] - yaxis[0])
            self.mask = np.copy(mask)
            self.mask[np.isnan(self.mask)] = 0
            self.normalized = normalized
            self.sparse_interp = sparse_interp
            self.intersections = Tree()
            self.ndv = ndv
            self.interp_points = InterpPoints()
            self.sup_interp_points = InterpPoints()
            self.dinterp = dinterp
            self.fz_old = None
            self.fz_new = None
            self.smooth = smooth
            self.surf_max = surf_max
            self.surf_min = surf_min

            self.contour = contour
            self.sections = sections

            if self.contour is None:
                self.set_contour()

            if len(self.sections) > 0:
                self.set_intersections()

            self.set_fz_old()

        except AssertionError as error:
            self.logger.error(error)
            raise

    @property
    def surf_min(self):
        return self._surf_min

    @surf_min.setter
    def surf_min(self, array):
        if array is None:
            self._surf_min = array
        else:
            if self.indexing == "raster":
                self._surf_min = np.flip(array, axis=1).T
            elif self.indexing == "ij":
                self._surf_min = array

    @surf_min.getter
    def surf_min(self):
        if self._surf_min is None:
            return self._surf_min
        else:
            if self.indexing == "raster":
                return np.flip(self._surf_min.T, axis=1)
            elif self.indexing == "ij":
                return self._surf_min

    @property
    def surf_max(self):
        return self._surf_max

    @surf_max.setter
    def surf_max(self, array):
        if array is None:
            self._surf_max = array
        else:
            if self.indexing == "raster":
                self._surf_max = np.flip(array, axis=1).T
            elif self.indexing == "ij":
                self._surf_max = array

    @surf_max.getter
    def surf_max(self):
        if self._surf_max is None:
            return self._surf_max
        else:
            if self.indexing == "raster":
                return np.flip(self._surf_max.T, axis=1)
            elif self.indexing == "ij":
                return self._surf_max

    @property
    def sections(self):
        return self._sections

    @sections.setter
    def sections(self, sections):
        if isinstance(sections, list):
            prof = []
            for section in sections:
                if isinstance(section, Section):
                    prof.append(section)
        elif isinstance(sections, str):
            if sections[-4:] == ".shp":
                prof = SurfMod.read_shapefile(sections)
            else:
                warnings.warn(
                    "sections initiation : file format not recognized"
                )
        self._sections = prof

        self.update_sections_extents()

        for section in self._sections:
            self[section.name] = section
            section.dinterp = self.dinterp

        for name1, name2 in itertools.product(self, self):
            for ff in ["points", "pos", "z"]:
                self.intersections[name1][name2][ff] = []

    @property
    def sections_extent(self):
        if len(self.sections) > 0:
            sec = self.sections[0]
            xx, yy = sec.line.xy
            xmin = min(xx)
            xmax = max(xx)
            ymin = min(yy)
            ymax = max(yy)
            for sec in self.sections[1:]:
                xx, yy = sec.line.xy
                xmin = min(xmin, min(xx))
                xmax = max(xmax, max(xx))
                ymin = min(ymin, min(yy))
                ymax = max(ymax, max(yy))
        else:
            xmin = None
            ymin = None
            xmax = None
            ymax = None
        return xmin, xmax, ymin, ymax

    def set_fz_old(self, finterp=None, outside_mask=False):
        if finterp is None:
            xmin, xmax, ymin, ymax = self.sections_extent
            indx = np.argwhere((self.xaxis >= xmin) & (self.xaxis <= xmax))[
                :, 0
            ]
            indy = np.argwhere((self.yaxis >= ymin) & (self.yaxis <= ymax))[
                :, 0
            ]
            xinterp = self.xaxis[indx[0] : indx[-1] + 1 : self.sparse_interp]
            yinterp = self.yaxis[indy[0] : indy[-1] + 1 : self.sparse_interp]
            zinterp = self.surf_old[
                indx[0] : indx[-1] + 1 : self.sparse_interp,
                indy[0] : indy[-1] + 1 : self.sparse_interp,
            ]
            if outside_mask:
                xinterp2, yinterp2 = np.meshgrid(
                    xinterp, yinterp, indexing="ij"
                )
                # Indexs inside the mask
                mask2 = self.mask[
                    indx[0] : indx[-1] + 1 : self.sparse_interp,
                    indy[0] : indy[-1] + 1 : self.sparse_interp,
                ]
                ind_mask = mask2 > 0
                # Indexes of mask contour
                mask3 = np.copy(mask2)
                mask3[ind_mask] = 1
                mask3 = convolve2d(
                    mask3,
                    np.ones((self.sparse_interp, self.sparse_interp)),
                    mode="same",
                )
                ind_cont = (mask3 > 0) & (mask2 == 0)
                # Interpoling function using contour
                fz_tmp = scipy.interpolate.Rbf(
                    xinterp2[ind_cont],
                    yinterp2[ind_cont],
                    zinterp[ind_cont],
                    function="multiquadric",
                    smooth=self.smooth,
                )
                # Modify altitudes within the mask for fz_old definition
                zinterp[ind_mask] = fz_tmp(
                    xinterp2[ind_mask], yinterp2[ind_mask]
                )
            self.fz_old = scipy.interpolate.RectBivariateSpline(
                xinterp, yinterp, zinterp
            )
        else:
            self.fz_old = finterp
        for section in self.sections:
            section.fz_old = self.fz_old
            if len(section.extent.pos) > 0:
                x0, y0 = section.extent.points.geoms[0].xy
                x1, y1 = section.extent.points.geoms[1].xy
                section.extent.z = [self.ndv, self.ndv]
                section.extent.z[0] = self.fz_old(x0[0], y0[0], grid=False)
                section.extent.z[-1] = self.fz_old(x1[0], y1[0], grid=False)

    def set_fz_new(self):
        if len(self.interp_points.z) > 0:
            self.fz_new = scipy.interpolate.Rbf(
                self.interp_points.xy[:, 0],
                self.interp_points.xy[:, 1],
                self.interp_points.z,
                function="multiquadric",
                smooth=self.smooth,
            )

    def set_surf_new(self):
        if self.fz_new is None:
            self.set_fz_new()
        if self.fz_new is not None:
            ind = self.mask > 0
            xmesh, ymesh = np.meshgrid(self.xaxis, self.yaxis, indexing="ij")
            self.surf_new[ind] = self.fz_new(xmesh[ind], ymesh[ind])
            if self.surf_min is not None:
                self.surf_new[ind] = np.maximum(
                    self.surf_new[ind], self.surf_min[ind]
                )
            if self.surf_max is not None:
                self.surf_new[ind] = np.minimum(
                    self.surf_new[ind], self.surf_max[ind]
                )

    def add_section(self, section):
        try:
            assert isinstance(
                section, Section
            ), "section is not a Section instance"
            self.sections.append(section)
            self[section.name] = section
            if self.contour is not None:
                self[section.name].set_extent(self.contour)
        except AssertionError:
            raise

    def update_sections_extents(self):
        if self.contour is not None:
            for section in self.sections:
                section.set_extent(self.contour)

    def set_contour(self, mask=None, x=None, y=None):
        plt.ioff()
        if mask is None:
            mask = self.mask
        if x is None:
            x = self.xaxis
        if y is None:
            y = self.yaxis

        try:
            backend = plt.get_backend()
            plt.switch_backend("Agg")
            fig, axe = plt.subplots(1, 1)
            cs = axe.contour(x, y, mask.T, [0.5])
            p = cs.collections[0].get_paths()
            assert len(p) > 0, "mask is either all 1 or all 0"
            p = p[0]
            self.contour = geom.Polygon(p.vertices)
            self.update_sections_extents()
            plt.switch_backend(backend)

        except ValueError:
            raise
        except TypeError:
            raise
        except AssertionError:
            raise
        plt.ion()

    def set_intersections(self, *args):
        names = [name for name in self]
        nsections = len(names)

        if len(args) == 0:
            for i in range(nsections):
                for j in range(i + 1, nsections):
                    intersect = self[names[i]].line.intersection(
                        self[names[j]].line
                    )
                    # print(intersect)
                    if isinstance(intersect, geom.Point):
                        intersect = geom.MultiPoint([intersect])
                    if isinstance(intersect, geom.LineString):
                        continue
                    self.intersections[names[i]][names[j]][
                        "points"
                    ] = intersect
                    self.intersections[names[j]][names[i]][
                        "points"
                    ] = intersect
                    posi = [
                        self[names[i]].line.project(
                            point, normalized=self.normalized
                        )
                        for point in intersect.geoms
                    ]
                    posj = [
                        self[names[j]].line.project(
                            point, normalized=self.normalized
                        )
                        for point in intersect.geoms
                    ]
                    self.intersections[names[i]][names[j]]["pos"] = posi
                    self.intersections[names[j]][names[i]]["pos"] = posj
                    self.intersections[names[i]][names[j]]["z"] = [
                        np.nan for point in intersect.geoms
                    ]
                    self.intersections[names[j]][names[i]]["z"] = [
                        np.nan for point in intersect.geoms
                    ]
        #                    print(names[i],names[j],posi)
        #                    print(names[j],names[i],posj)
        else:
            names1 = args[0]
            names2 = args[1]
            zs = args[2]
            assert len(names1) == len(names2)
            assert len(names1) == len(zs)
            for key1, key2, z in zip(names1, names2, zs):
                if (key1 in names) and (key2 in names):
                    self.intersections[key1][key2]["z"] = [z]
                    self.intersections[key2][key1]["z"] = [z]

                    self[key1].set_intersection_points(
                        key2,
                        self.intersections[key1][key2]["pos"],
                        self.intersections[key1][key2]["z"],
                    )
                    self[key2].set_intersection_points(
                        key1,
                        self.intersections[key2][key1]["pos"],
                        self.intersections[key2][key1]["z"],
                    )

    def set_control_points(self, *args, sep=" "):
        if len(args) == 2:
            pos = args[0]
            z = args[1]
        elif len(args) == 1:
            (
                pos,
                z,
                z_intersections,
            ) = SurfMod.read_control_points_from_textfile(args[0], sep=sep)
            for name1 in z_intersections:
                for name2 in z_intersections[name1]:
                    zz = z_intersections[name1][name2]
                    self.intersections[name1][name2]["z"] = [zz]
                    self.intersections[name2][name1]["z"] = [zz]
        else:
            pos = []
            z = []

        if (isinstance(pos, dict)) & (isinstance(z, dict)):
            for key in self:
                if key in pos:
                    self[key].set_control_points(pos[key], z[key])
                for key2 in self.intersections[key]:
                    self[key].set_intersection_points(
                        key2,
                        self.intersections[key][key2]["pos"],
                        self.intersections[key][key2]["z"],
                    )
                #                        print(key,key2,self.intersections[key][key2]['pos'],self.intersections[key][key2]['z'])
                self[key].filter_control_points()

        else:
            raise ValueError("set_control_points recieved incorrect arguments")

    def set_interp_points(self):
        interp_points = []
        interp_z = []
        if self.contour is None:
            self.set_contour()

        # Add points on the contour
        if self.contour is not None:
            exterior = self.contour.exterior
            npts = int(np.floor(exterior.length / self.dinterp) + 1)
            ss = np.linspace(0, exterior.length, npts)
            new_pts = [exterior.interpolate(s) for s in ss]
            interp_points = interp_points + new_pts
            coords = [pt.coords[:][0] for pt in new_pts]
            interp_z = interp_z + [
                self.fz_old(coord[0], coord[1], grid=False) for coord in coords
            ]

        # Add points from the sections
        for section in self.sections:
            section.set_interp_spline()
            section.set_interp_points()
            interp_points = interp_points + [
                pp for pp in section.interp_points.points.geoms
            ]
            interp_z = interp_z + section.interp_points.z

        # Add supplementary control points
        interp_points = interp_points + [
            pp for pp in self.sup_interp_points.points.geoms
        ]
        interp_z = interp_z + self.sup_interp_points.z

        self.interp_points.points = geom.MultiPoint(interp_points)
        self.interp_points.z = interp_z
        self.interp_points.unique()

    def update_interpolation(self):
        self.set_interp_points()
        self.set_fz_new()
        self.set_surf_new()

    def set_sup_interp_points(self, *args):
        self.sup_interp_points = InterpPoints(*args)

    def add_sup_interp_points(self, *args):
        self.sup_interp_points.add(*args)

    def plot(
        self,
        figsize=None,
        axe=None,
        contour_properties=None,
        section_properties=None,
        text_properties=None,
        control_points_properties=None,
        plot_interp_points=False,
        interp_points_properties=None,
        point_name_dist=None,
        npts=2,
        **kwargs,
    ):
        if axe is None:
            fig, axe = plt.subplots(1, 1, figsize=figsize)
        else:
            fig = axe.figure

        # Plot topography

        digdem.plot.plot_topo(
            self.surf_new,
            self.xaxis,
            self.yaxis,
            indexing=self.indexing,
            axe=axe,
            **kwargs,
        )

        # Plot contour
        if contour_properties is None:
            contour_properties = dict()
        xc, yc = self.contour.exterior.coords.xy
        axe.plot(xc, yc, **contour_properties)

        # Plot sections
        if control_points_properties is None:
            control_points_properties = {
                "marker": "o",
                "ms": 5,
                "mec": "white",
                "mfc": "k",
            }
        for section in self.sections:
            section.plot_on_map(
                axe=axe,
                section_properties=section_properties,
                text_properties=text_properties,
                control_points_properties=control_points_properties,
                npts=npts,
                point_name_dist=point_name_dist,
            )

        for pt in self.sup_interp_points.points.geoms:
            axe.plot(
                pt.coords[0][0], pt.coords[0][1], **control_points_properties
            )

        if plot_interp_points:
            for pt in self.interp_points.points.geoms:
                axe.plot(
                    pt.coords[0][0],
                    pt.coords[0][1],
                    **interp_points_properties,
                )

        return fig, axe


if __name__ == "__main__":
    folder_shapefiles = (
        "/home/peruzzetto/Documents/martinique/qgis/shapefiles/"
    )
    shapefile = os.path.join(folder_shapefiles, "cliff_sections.shp")
    surfmod = SurfMod(np.ones((2, 3)), mask=np.ones((2, 3)))



#!/localstorage/peruzzetto/anaconda3/bin/python
# -*- coding: utf-8 -*-
"""
Created on Tue Feb  4 11:49:44 2020

@author: peruzzetto
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

BOLD_CONTOURS_INTV = [0.1, 0.2, 0.5, 1, 2.0, 5, 10, 20, 50, 100, 200, 500, 1000]
NB_THIN_CONTOURS = 10
NB_BOLD_CONTOURS = 3


def get_contour_intervals(zmin, zmax, nb_bold_contours=None, nb_thin_contours=None):
    if nb_thin_contours is None:
        nb_thin_contours = NB_THIN_CONTOURS
    if nb_bold_contours is None:
        nb_bold_contours = NB_BOLD_CONTOURS

    intv = (zmax - zmin) / nb_bold_contours
    i = np.argmin(np.abs(np.array(BOLD_CONTOURS_INTV) - intv))

    bold_intv = BOLD_CONTOURS_INTV[i]
    if BOLD_CONTOURS_INTV[i] != BOLD_CONTOURS_INTV[0]:
        if bold_intv - intv > 0:
            bold_intv = BOLD_CONTOURS_INTV[i - 1]

    if nb_thin_contours is None:
        thin_intv = bold_intv / NB_THIN_CONTOURS
        if (zmax - zmin) / bold_intv > 5:
            thin_intv = thin_intv * 2
    else:
        thin_intv = bold_intv / nb_thin_contours

    return bold_intv, thin_intv


def plot_topo(
    z,
    x,
    y,
    indexing='ij',
    contour_step=None,
    nlevels=None,
    level_min=None,
    step_contour_bold="auto",
    contour_labels_properties=None,
    label_contour=True,
    contour_label_effect=None,
    axe=None,
    vert_exag=1,
    fraction=1,
    ndv=-9999,
    uniform_grey=None,
    contours_prop=None,
    contours_bold_prop=None,
    figsize=None,
    interpolation=None,
    sea_level=0,
    sea_color=None,
    alpha=1,
    azdeg=315,
    altdeg=45,
    zmin=None,
    zmax=None,
):
    """
    Plot topography with hillshading.

    Parameters
    ----------
    z : TYPE
        DESCRIPTION.
    x : TYPE
        DESCRIPTION.
    y : TYPE
        DESCRIPTION.
    contour_step : TYPE, optional
        DESCRIPTION. The default is None.
    nlevels : TYPE, optional
        DESCRIPTION. The default is 25.
    contour_labels_properties : TYPE, optional
        DESCRIPTION. The default is None.
    axe : TYPE, optional
        DESCRIPTION. The default is None.
    vert_exag : TYPE, optional
        DESCRIPTION. The default is 1.
    contour_color : TYPE, optional
        DESCRIPTION. The default is 'k'.
    contour_alpha : TYPE, optional
        DESCRIPTION. The default is 1.
    figsize : TYPE, optional
        DESCRIPTION. The default is (10,10).
    interpolation : TYPE, optional
        DESCRIPTION. The default is None.
    sea_level : TYPE, optional
        DESCRIPTION. The default is 0.
    sea_color : TYPE, optional
        DESCRIPTION. The default is None.
    alpha : TYPE, optional
        DESCRIPTION. The default is 1.
    azdeg : TYPE, optional
        DESCRIPTION. The default is 315.
    altdeg : TYPE, optional
        DESCRIPTION. The default is 45.
    fraction : TYPE, optional
        DESCRIPTION. The default is 1.
    ndv : TYPE, optional
        DESCRIPTION. The default is -9999.

    Returns
    -------
    None.

    """
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    im_extent = [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]
    ls = mcolors.LightSource(azdeg=azdeg, altdeg=altdeg)
    
    if indexing == 'ij':
        z = np.flip(z, axis=1).T

    auto_bold_intv = None

    if nlevels is None and contour_step is None:
        auto_bold_intv, contour_step = get_contour_intervals(np.nanmin(z), np.nanmax(z))

    if level_min is None:
        if contour_step is not None:
            level_min = np.ceil(np.nanmin(z) / contour_step) * contour_step
        else:
            level_min = np.nanmin(z)
    if contour_step is not None:
        levels = np.arange(level_min, np.nanmax(z), contour_step)
    else:
        levels = np.linspace(level_min, np.nanmax(z), nlevels)

    if axe is None:
        fig = plt.figure(figsize=figsize)
        axe = fig.gca()
    else:
        fig = axe.figure
    axe.set_ylabel("Y (m)")
    axe.set_xlabel("X (m)")
    axe.set_aspect("equal")

    if uniform_grey is None:
        shaded_topo = ls.hillshade(z, vert_exag=vert_exag, dx=dx, dy=dy, fraction=1)
    else:
        shaded_topo = np.ones(z.shape) * uniform_grey
    shaded_topo[z == ndv] = np.nan
    axe.imshow(
        shaded_topo,
        cmap="gray",
        extent=im_extent,
        interpolation=interpolation,
        alpha=alpha,
        vmin=0,
        vmax=1,
    )

    if contours_prop is None:
        contours_prop = dict(alpha=0.5, colors="k", linewidths=0.5)
    axe.contour(
        x, y, np.flip(z, axis=0), extent=im_extent, levels=levels, **contours_prop
    )

    if contours_bold_prop is None:
        contours_bold_prop = dict(alpha=0.8, colors="k", linewidths=0.8)

    if step_contour_bold == "auto":
        if auto_bold_intv is None:
            auto_bold_intv, _ = get_contour_intervals(np.nanmin(z), np.nanmax(z))
        step_contour_bold = auto_bold_intv

    if step_contour_bold > 0:
        lmin = np.ceil(np.nanmin(z) / step_contour_bold) * step_contour_bold
        if lmin < level_min:
            lmin = lmin + step_contour_bold
        levels = np.arange(lmin, np.nanmax(z), step_contour_bold)
        cs = axe.contour(
            x,
            y,
            np.flip(z, axis=0),
            extent=im_extent,
            levels=levels,
            **contours_bold_prop,
        )
        if label_contour:
            if contour_labels_properties is None:
                contour_labels_properties = {}
            clbls = axe.clabel(cs, **contour_labels_properties)
            if contour_label_effect is not None:
                plt.setp(clbls, path_effects=contour_label_effect)

    if sea_color is not None:
        cmap_sea = mcolors.ListedColormap([sea_color])
        cmap_sea.set_under(color="w", alpha=0)
        mask_sea = (z <= sea_level) * 1
        if mask_sea.any():
            axe.imshow(
                mask_sea,
                extent=im_extent,
                cmap=cmap_sea,
                vmin=0.5,
                origin="lower",
                interpolation="none",
            )
